! $Id: PolyScalarSharedWeakPointerExample.f90 2220 2016-06-23 00:17:02Z ian $
! Copyright 2016 M.E.G.M.S.  See LICENCE.txt for licence (Apache 2.0).
! See http://www.megms.com.au/shared-pointers.htm for more information.

! Example and simple tests of the code in PolyScalarSharedWeakPointer.i90.

! A module that defines the type that the shared pointer will 
! reference, and also brings in the template code to define the 
! shared pointer.
!
! The object type includes the capability for circular references.

MODULE MyTypes_implementation
  IMPLICIT NONE
  
  ! We could make the default access PRIVATE and explicitly list 
  ! all the relevant identifiers as PUBLIC, but as this is 
  ! an internal implementation module we don't bother - the control 
  ! of identifiers for client code is left to the `MyTypes` module.
  
  ! Incomplete F2008 support means that allocatable components cannot 
  ! forward reference a type, hence we have to use a pointer component.  
  ! This means we want finalization to ensure the pointer component 
  ! is deallocated when appropriate, but you cannot have a finalizer 
  ! for an abstract type, so we split the pointer component out into 
  ! another non-abstract parent type.
  TYPE :: parent_impl
    TYPE(SharedPointer), POINTER :: next => NULL()
    ! As above, but using the WeakPointer reference type.
    TYPE(WeakPointer), POINTER :: weak_next => NULL()
  CONTAINS
    FINAL :: pi_final
  END TYPE parent_impl
  
  ! The parent type.  This will be used as the object_type.  Because 
  ! we are defining this type in the same module that defines the 
  ! shared pointer, we need to give it the name `object_type`.  The 
  ! type will be renamed in a later module.
  TYPE, ABSTRACT, EXTENDS(parent_impl) :: object_type
  CONTAINS
    ! Example binding.
    PROCEDURE(parent_Binding), DEFERRED :: Binding
  END TYPE object_type
  
  ABSTRACT INTERFACE
    ! Interface for the example binding in object_type.
    SUBROUTINE parent_Binding(arg)
      IMPORT :: object_type
      CLASS(object_type), INTENT(IN) :: arg
    END SUBROUTINE parent_Binding
  END INTERFACE
  
  ! Bring in the template code that defines the shared pointer, 
  ! the weak pointer and the relevant procedures.
  INCLUDE 'PolyScalarSharedWeakPointer.i90'
  
  SUBROUTINE pi_final(obj)
    TYPE(parent_impl), INTENT(INOUT) :: obj
    
    IF (ASSOCIATED(obj%next)) THEN
      DEALLOCATE(obj%next)
    END IF
  END SUBROUTINE pi_final
END MODULE MyTypes_implementation


! Module that renames the relevant entities from 
! MyTypes_implementation.

MODULE MyTypes
  ! Rename object_type to the name that we actually want client 
  ! code to use.  We could also rename the shared and weak pointer 
  ! type names here too.
  USE MyTypes_implementation, MyParent => object_type
  
  IMPLICIT NONE
  
  PRIVATE
  
  ! Identifiers from MyTypes_implementation that we want to be visible 
  ! in client code.
  PUBLIC :: MyParent
  PUBLIC :: SharedPointer
  PUBLIC :: WeakPointer
  PUBLIC :: Acquire
  PUBLIC :: Get
  PUBLIC :: IsAssociated
  
  ! An extension of MyParent, that will be used as the dynamic 
  ! type of the object that the shared pointer references.
  TYPE, EXTENDS(MyParent), PUBLIC :: MyExtension
    CHARACTER(:), ALLOCATABLE :: comp
  CONTAINS
    PROCEDURE :: Binding => extension_Binding
    FINAL :: extension_Final
  END TYPE MyExtension
CONTAINS
  SUBROUTINE extension_Binding(arg)
    CLASS(MyExtension), INTENT(IN) :: arg
    PRINT "(2X,'Binding for ',A)", arg%comp
  END SUBROUTINE extension_Binding
  
  IMPURE ELEMENTAL SUBROUTINE extension_Final(arg)
    TYPE(MyExtension), INTENT(INOUT) :: arg
    PRINT "(2X,'The object for ',A,' is going away.')", arg%comp
  END SUBROUTINE extension_Final
END MODULE MyTypes


! Example/test driver.

PROGRAM PolyScalarExample
  USE MyTypes
  IMPLICIT NONE
  
  ! Procedures that demonstrate/test use of the shared pointer.
  CALL test01
  CALL test02
  CALL test03
  CALL test04
  CALL test05
  CALL test06
  CALL test07
  CALL test08
  
  ! These procedures show how you can break the shared pointer.
!  CALL evil01
!  CALL evil02
!  CALL evil03
CONTAINS
  ! Basic tests using Acquire and ptr%Item.
  SUBROUTINE test01
    TYPE(SharedPointer) :: a, b
    
    PRINT "('test01')"
    
    ! Before definition, a shared pointer references no object.
    PRINT "(2X,'Does a reference an object: ',L1)", IsAssociated(a)
    
    BLOCK
      CLASS(MyParent), POINTER :: object
      ! Set up the initial object.
      PRINT "(2X,'Creating Red')"
      ALLOCATE(object, SOURCE=MyExtension(NULL(), NULL(), 'Red'))
      ! Hand control of the lifetime of that object to a shared pointer.
      CALL Acquire(a, object)
      
      ! After calling Acquire, the lifetime of the target of the 
      ! Fortran pointer is managed by the shared pointer.  Consistent 
      ! with this, the association status of the Fortran pointer is 
      ! set to dissociated.
      PRINT "(2X,'The association status of object is ',L1)",  &
          ASSOCIATED(object)
      PRINT "(2X,'Does a reference an object: ',L1)",  &
          IsAssociated(a)
    END BLOCK
    
    ! The Item component can be used to work with the referenced 
    ! object.  This should print `Red`.
    CALL a%Item%Binding()
    
    ! Assignment can be used to create additional references to 
    ! the same object.
    b = a
    CALL b%Item%Binding()
    
    ! A shared pointer can be reset by assigning an empty pointer 
    ! to it.
    a = SharedPointer()
    
    ! The object that was referenced by `a` persists, because 
    ! `b` also references it.
    !
    ! The `b` shared pointer still references the object, but when 
    ! `b` goes out of scope at the end of the procedure the object 
    ! will be deallocated.
    PRINT "(2X,'end test01')"
  END SUBROUTINE test01
  
  ! Test SharedPointer constructor and Get.
  SUBROUTINE test02
    TYPE(SharedPointer) :: a, b
    CLASS(MyParent), POINTER :: p
    
    PRINT "('test02')"
    
    BLOCK
      CLASS(MyParent), POINTER :: object
      ! Set up the initial object.
      PRINT "(2X,'Creating Blue')"
      ALLOCATE(object, SOURCE=MyExtension(NULL(), NULL(), 'Blue'))
      ! Hand control of the lifetime of that object to a shared pointer.  
      ! This time use the overloaded structure constructor.
      a = SharedPointer(object)
    END BLOCK
    
    p => Get(a)
    CALL p%Binding()
    
    ! Function that implements the BLOCK above.
    b = make_ptr('Green')
    
    ! Green will not be referenced by anything after this assignment, 
    ! so it will be destroyed.  a will reference Red.
    a = b
    
    p => Get(a)
    CALL p%Binding()
    
    PRINT "(2X,'end test02')"
  END SUBROUTINE test02
  
  ! Like test02, but with shared pointer inside other type.
  SUBROUTINE test03
    TYPE :: t
      TYPE(SharedPointer) :: sp_comp
    END TYPE t
    TYPE(t) :: x, y
    CLASS(MyParent), POINTER :: p
    
    PRINT "('test03')"
    
    ! Point x%sp_comp at Blue.
    x%sp_comp = make_ptr('Blue')
    
    p => Get(x%sp_comp)
    CALL p%Binding()
    
    ! Point y%sp_comp at Green.
    y%sp_comp = make_ptr('Green')
    
    x = y
    
    p => Get(x%sp_comp)
    CALL p%Binding()
    PRINT "(2X,'end test03')"
  END SUBROUTINE test03
  
  ! Test passing to procedure with VALUE attribute.
  SUBROUTINE test04
    TYPE(SharedPointer) :: a
    
    PRINT "('test04')"
    
    a = make_ptr('Blue')
    
    PRINT "(2X,'Before worker')"
    CALL test04_worker(a)
    PRINT "(2X,'After worker')"
    PRINT "(2X,'end test04')"
  END SUBROUTINE test04
  
  ! Worker procedure for test04.  Takes shared pointer argument by 
  ! value.
  SUBROUTINE test04_worker(x)
    TYPE(SharedPointer), VALUE :: x
    
    ! arg is not finalized.
  END SUBROUTINE test04_worker
  
  ! Allocatable shared pointer.
  SUBROUTINE test05
    TYPE(SharedPointer) :: a
    TYPE(SharedPointer), ALLOCATABLE :: b
    
    PRINT "('test05')"
    
    a = make_ptr('Blue')
    
    ! This is intrinsic assignment.
    b = a
    
    a = SharedPointer()
    
    DEALLOCATE(b)
    
    PRINT "(2X,'end test05')"
  END SUBROUTINE test05
  
  ! Simple weak pointer tests.
  SUBROUTINE test06
    TYPE(SharedPointer) :: a, c
    TYPE(WeakPointer) :: weak
    
    PRINT "('test06')"
    
    a = make_ptr('Blue')
    
    ! Get a weak pointer
    weak = WeakPointer(a)
    
    ! Get a shared pointer from that weak pointer.
    c = SharedPointer(weak)
    PRINT "(2X,'Association status of c%Item is ',L1)",  &
        ASSOCIATED(c%Item)
    PRINT "(2X,'resetting c')"
    c = SharedPointer()
    
    PRINT "(2X,'resetting a')"
    a = SharedPointer()     ! Blue should be deallocated.
    
    ! Get a shared pointer now that the referenced object has gone.
    c = SharedPointer(weak)
    PRINT "(2X,'Association status of c%Item is ',L1)",  &
        ASSOCIATED(c%Item)
    
    PRINT "(2X,'resetting weak')"
    weak = WeakPointer()    ! Control block should be deallocated.
    
    ! Get a shared pointer from a weak pointer that references 
    ! no object.
    a = SharedPointer(weak)
    PRINT "(2X,'Association status of a%Item is ',L1)",  &
        ASSOCIATED(a%Item)
    
    ! Get a weak pointer from a shared pointer that references 
    ! no object.
    weak = WeakPointer(a)
    
    ! Get a shared pointer from a weak pointer that references 
    ! no object (again).
    a = SharedPointer(weak)
    PRINT "(2X,'Association status of a%Item is ',L1)",  &
        ASSOCIATED(a%Item)
    
    PRINT "(2X,'end test06')"
  END SUBROUTINE test06
  
  ! Create and destroy weak pointers before all shared pointers.
  SUBROUTINE test07
    TYPE(SharedPointer) :: a
    TYPE(WeakPointer) :: wa, wb
    
    PRINT "('test07')"
    
    a = make_ptr('Blue')
    
    wa = WeakPointer(a)
    wb = wa
    
    ! Reset the weak pointers.
    PRINT "(2X,'resetting wa')"
    wa = WeakPointer()
    PRINT "(2X,'resetting wb')"
    wb = WeakPointer()
    
    ! Reset the two local shared pointer objects.
    PRINT "(2X,'resetting a')"
    a = SharedPointer()   ! Blue should be deallocated.
    
    PRINT "(2X,'end test07')"
  END SUBROUTINE test07
  
  ! Have weak pointers outlive shared pointers.
  SUBROUTINE test08
    TYPE(SharedPointer) :: a
    TYPE(WeakPointer) :: wa, wb
    
    PRINT "('test08')"
    
    a = make_ptr('Blue')
    
    wa = WeakPointer(a)
    wb = wa
    
    ! Reset the two local shared pointer objects.
    PRINT "(2X,'resetting a')"
    a = SharedPointer()
    
    ! Reset the weak pointers.
    PRINT "(2X,'resetting wa')"
    wa = WeakPointer()
    PRINT "(2X,'resetting wb')"
    wb = WeakPointer()    ! Control block should be deallocated.
    
    PRINT "(2X,'end test08')"
  END SUBROUTINE test08
  
  ! Create a cycle using weak pointers
  SUBROUTINE test09
    TYPE(SharedPointer) :: a, b
    
    PRINT "('test09')"
    
    a = make_ptr('Blue')
    b = make_ptr('Green')
    
    ! Allocate shared pointer components and create cycle.
    ALLOCATE(a%item%weak_next)
    a%item%weak_next = WeakPointer(b)
    ALLOCATE(b%item%weak_next)
    b%item%weak_next = WeakPointer(a)
    
    ! Reset the two local shared pointer objects.
    a = SharedPointer()   ! Blue should be deallocated.
    b = SharedPointer()   ! Green should be deallocated.
    
    PRINT "(2X,'end test09')"
  END SUBROUTINE test09
  
  ! Break using VALUE attribute.
  SUBROUTINE evil01
    TYPE(SharedPointer) :: a
    
    PRINT "('evil01')"
    
    a = make_ptr('Blue')
    
    PRINT "(2X,'Before worker')"
    CALL evil01_worker(a)
    PRINT "(2X,'After worker')"
    PRINT "(2X,'end evil01')"
  END SUBROUTINE evil01
  
  ! Worker procedure for evil01.  Takes shared pointer argument by 
  ! value, then modifies the shared pointer.
  SUBROUTINE evil01_worker(x)
    TYPE(SharedPointer), VALUE :: x    ! Not INTENT(IN).
    
    ! Vary x by pointing it at another object.
    x = make_ptr('Green')
    
    ! x not finalized here - memory leak.
  END SUBROUTINE evil01_worker
  
  ! Break using ALLOCATE(..., SOURCE=xxx)
  SUBROUTINE evil02
    TYPE(SharedPointer) :: a
    TYPE(SharedPointer), ALLOCATABLE :: b
    
    PRINT "('evil02')"
    
    a = make_ptr('Blue')
    
    ! Does not update reference count.
    ALLOCATE(b, SOURCE=a)
    
    ! Reference count will be zero - object and control block deleted.
    a = SharedPointer()
    
    ! Finalization of b will attempt to reference a control block
    ! that has already gone.  This is non-conforming, may or may 
    ! not be identified at runtime.
    PRINT "(2X,'end evil02')"
  END SUBROUTINE evil02
  
  ! Break by creating a cycle.
  SUBROUTINE evil03
    TYPE(SharedPointer) :: a, b
    
    PRINT "('evil03')"
    
    a = make_ptr('Blue')
    b = make_ptr('Green')
    
    ! Allocate shared pointer components and create cycle.
    ALLOCATE(a%item%next)
    a%item%next = b
    ALLOCATE(b%item%next)
    b%item%next = a
    
    ! Reset the two local shared pointer objects.
    a = SharedPointer()
    b = SharedPointer()
    
    ! Memory leak here - neither Blue or Green are cleared up.
    PRINT "(2X,'end evil03')"
  END SUBROUTINE evil03
  
  ! Utility to allocate a target and place it under the control 
  ! of a shared pointer.
  FUNCTION make_ptr(colour) RESULT(ptr)
    CHARACTER(*), INTENT(IN) :: colour
    TYPE(SharedPointer) :: ptr
    CLASS(MyParent), POINTER :: object
    
    PRINT "(2X,'Creating ',A)", colour
    ALLOCATE(object, SOURCE=MyExtension(NULL(), NULL(), colour))
    ptr = SharedPointer(object)
  END FUNCTION make_ptr
END PROGRAM PolyScalarExample

